# Copyright 2021 CR-Suite Development Team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cr.nimble._src.subspaces import (
    orth_complement,
    project_to_subspace,
    is_in_subspace,
    principal_angles_cos,
    principal_angles_cos_jit,
    principal_angles_rad,
    principal_angles_rad_jit,
    principal_angles_deg,
    principal_angles_deg_jit,
    smallest_principal_angle_cos,
    smallest_principal_angle_cos_jit,
    smallest_principal_angle_rad,
    smallest_principal_angle_rad_jit,
    smallest_principal_angle_deg,
    smallest_principal_angle_deg_jit,
    smallest_principal_angles_cos,
    smallest_principal_angles_cos_jit,
    smallest_principal_angles_rad,
    smallest_principal_angles_rad_jit,
    smallest_principal_angles_deg,
    smallest_principal_angles_deg_jit,
    subspace_distance,
)